package main

import (
	"github.com/gin-gonic/gin"
	log "github.com/sirupsen/logrus"
	"io/ioutil"
	"net/http"
	"random.chars.jp/git/image-board/api"
	"random.chars.jp/git/image-board/store"
	"strconv"
	"strings"
	"unicode/utf8"
)

func registerAPI() {
	router.GET(api.Base, func(context *gin.Context) {
		context.JSON(http.StatusOK, store.Info{
			Revision:       instance.Revision,
			Compat:         instance.Compat,
			Register:       instance.Register,
			InitialUser:    instance.InitialUser,
			PermissionDir:  instance.PermissionDir,
			PermissionFile: instance.PermissionFile,
		})
	})

	router.GET(api.SingleUser, func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.SingleUser)
	})

	router.GET(api.Private, func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.Private)
	})

	router.GET(api.User, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.JSON(http.StatusOK, instance.Users())
	})

	router.PUT(api.User, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		user, ok := getUser(context)
		if !instance.Register {
			if !ok {
				context.JSON(http.StatusForbidden, api.Error{Error: "user registration disallowed"})
				return
			}

			if !user.Privileged {
				context.JSON(http.StatusForbidden, api.Denied)
				return
			}
		}

		var payload api.UserCreatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}
		if !user.Privileged && payload.Privileged {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}

		context.JSON(http.StatusOK, instance.UserAdd(payload.Username, payload.Password, payload.Privileged))
	})

	router.GET(api.UserThis, func(context *gin.Context) {
		info, ok := getUser(context)
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}
		context.JSON(http.StatusOK, api.UserPayload{
			Username:   info.Username,
			ID:         info.Snowflake,
			Privileged: info.Privileged,
		})
	})

	router.GET(api.UserField, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		info := instance.User(context.Param("flake"))
		context.JSON(http.StatusOK, api.UserPayload{
			Username:   info.Username,
			ID:         info.Snowflake,
			Privileged: info.Privileged,
		})
	})

	router.PATCH(api.UserField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}

		var payload api.UserUpdatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}

		instance.UserUsernameUpdate(flake, payload.Username)
	})

	router.DELETE(api.UserField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		instance.UserDestroy(flake)
	})

	router.DELETE(api.UserPassword, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require privileged
		if !ok || !info.Privileged {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		instance.UserPasswordUpdate(flake, "")
		instance.UserSecretRegen(flake)
	})

	router.PUT(api.UserPassword, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, api.Denied)
		}

		var newPass string
		if payload, err := context.GetRawData(); err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		} else {
			if !utf8.Valid(payload) {
				context.JSON(http.StatusBadRequest, api.Error{Error: "invalid encoding"})
				return
			}
			newPass = string(payload)
			if len(newPass) > 8192 || strings.Contains(newPass, "\n") {
				context.JSON(http.StatusBadRequest, api.Error{Error: "invalid password"})
				return
			}
		}

		if newPass == "" {
			context.JSON(http.StatusBadRequest, api.Error{Error: "empty passwords are not allowed"})
			return
		}

		instance.UserPasswordUpdate(info.Snowflake, newPass)
		context.JSON(http.StatusOK, api.UserSecretPayload{Secret: instance.UserSecretRegen(info.Snowflake)})
	})

	router.GET(api.UsernameField, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		info := instance.UserUsername(context.Param("name"))
		payload := api.UserPayload{
			Username:   info.Username,
			ID:         info.Snowflake,
			Privileged: info.Privileged,
		}
		context.JSON(http.StatusOK, payload)
	})

	router.POST(api.UsernameAuth, func(context *gin.Context) {
		var password string

		if payload, err := context.GetRawData(); err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		} else {
			password = string(payload)
		}

		username := context.Param("name")
		if instance.UserUsernamePasswordValidate(username, password) {
			context.JSON(http.StatusOK, api.UserSecretPayload{Secret: instance.UserUsername(username).Secret})
		} else {
			context.JSON(http.StatusForbidden, api.Denied)
		}
	})

	router.GET(api.UserSecret, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		// Only allow lookup if user is current user or privileged
		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		context.JSON(http.StatusOK, api.UserSecretPayload{Secret: instance.User(flake).Secret})
	})

	router.PUT(api.UserSecret, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		// Only allow set if user is current user or privileged
		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		context.JSON(http.StatusOK, api.UserSecretPayload{Secret: instance.UserSecretRegen(flake)})
	})

	router.GET(api.UserImage, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.JSON(http.StatusOK, instance.UserImages(context.Param("flake")))
	})

	router.GET(api.SearchField, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		tagsPayload := context.Param("tags")
		tags := strings.Split(tagsPayload, "!")
		context.JSON(http.StatusOK, instance.ImageSearch(tags))
	})

	router.GET(api.Image, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require privileged
		if !ok || !info.Privileged {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		context.JSON(http.StatusOK, instance.ImageSnowflakes())
	})

	router.POST(api.Image, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		payload, err := context.FormFile("image")
		if err != nil {
			context.JSON(http.StatusInternalServerError, api.Error{Error: err.Error()})
			return
		}
		file, err := payload.Open()
		if err != nil {
			log.Errorf("Error while opening uploaded file %s, %s", payload.Filename, err)
			context.JSON(http.StatusInternalServerError, api.Error{Error: err.Error()})
			return
		}
		data, err := ioutil.ReadAll(file)
		if err != nil {
			log.Errorf("Error while reading uploaded file %s, %s", payload.Filename, err)
			context.JSON(http.StatusInternalServerError, api.Error{Error: err.Error()})
			return
		}
		image := instance.ImageAdd(data, info.Snowflake)
		if image.Hash == "" {
			context.JSON(http.StatusBadRequest, api.Error{Error: "invalid image"})
			return
		}
		context.JSON(http.StatusOK, image)
	})

	router.GET(api.ImagePage, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.String(http.StatusOK, strconv.Itoa(instance.PageTotal(store.ImageRootPageVariant)))
	})

	router.GET(api.ImagePageField, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.Page(store.ImageRootPageVariant, entry))
	})

	router.GET(api.ImagePageImage, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.PageImages(store.ImageRootPageVariant, entry))
	})

	router.GET(api.ImageField, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.JSON(http.StatusOK, instance.ImageSnowflake(context.Param("flake")))
	})

	router.PATCH(api.ImageField, func(context *gin.Context) {
		user, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		info := instance.ImageSnowflake(context.Param("flake"))
		if !user.Privileged && (info.User != user.Snowflake) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}

		var payload api.ImageUpdatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}

		instance.ImageUpdate(info.Hash, payload.Source, payload.Parent, payload.Commentary, payload.CommentaryTranslation)
	})

	router.DELETE(api.ImageField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		image := instance.ImageSnowflake(flake)
		if !info.Privileged && (info.Snowflake != image.User) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		instance.ImageDestroy(image.Hash)
	})

	router.GET(api.ImageFile, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		flake := context.Param("flake")
		image, data := instance.ImageData(instance.ImageSnowflakeHash(flake), false)
		if image.Snowflake != flake {
			context.JSON(http.StatusNotFound, api.Error{Error: "not found"})
			return
		}
		context.Data(http.StatusOK, "image/"+image.Type, data)
	})

	router.GET(api.ImagePreview, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		flake := context.Param("flake")
		image, data := instance.ImageData(instance.ImageSnowflakeHash(flake), true)
		if image.Snowflake != flake {
			context.JSON(http.StatusNotFound, api.Error{Error: "not found"})
			return
		}
		context.Data(http.StatusOK, "image/jpeg", data)
	})

	router.GET(api.ImageTag, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.JSON(http.StatusOK, instance.ImageTags(context.Param("flake")))
	})

	router.PUT(api.ImageTagField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != instance.ImageSnowflake(flake).User) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		instance.ImageTagAdd(flake, context.Param("tag"))
	})

	router.DELETE(api.ImageTagField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != instance.ImageSnowflake(flake).User) {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		instance.ImageTagRemove(flake, context.Param("tag"))
	})

	router.GET(api.Tag, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.JSON(http.StatusOK, instance.Tags())
	})

	router.GET(api.TagField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		context.JSON(http.StatusOK, instance.Tag(context.Param("tag")))
	})

	router.PUT(api.TagField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		context.JSON(http.StatusOK, instance.TagCreate(context.Param("tag")))
	})

	router.DELETE(api.TagField, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}
		instance.TagDestroy(context.Param("tag"))
	})

	router.GET(api.TagPage, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		tag := context.Param("tag")
		if !instance.MatchName(tag) {
			context.JSON(http.StatusBadRequest, api.Denied)
			return
		}
		context.String(http.StatusOK, strconv.Itoa(instance.PageTotal("tag_"+tag)))
	})

	router.GET(api.TagPageField, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		tag := context.Param("tag")
		if !instance.MatchName(tag) {
			context.JSON(http.StatusBadRequest, api.Denied)
			return
		}

		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.Page("tag_"+tag, entry))
	})

	router.GET(api.TagPageImage, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		tag := context.Param("tag")
		if !instance.MatchName(tag) {
			context.JSON(http.StatusBadRequest, api.Denied)
			return
		}

		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.PageImages("tag_"+tag, entry))
	})

	router.GET(api.TagInfo, func(context *gin.Context) {
		if !privateAccessible(context) {
			return
		}

		context.JSON(http.StatusOK, instance.TagInfo(context.Param("tag")))
	})

	router.PATCH(api.TagInfo, func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, api.Unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, api.Denied)
			return
		}

		var payload api.TagUpdatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, api.Error{Error: err.Error()})
			return
		}
		instance.TagType(context.Param("tag"), payload.Type)
	})
}

func privateAccessible(context *gin.Context) bool {
	if !instance.Private {
		return true
	}

	if _, ok := getUser(context); !ok {
		context.JSON(http.StatusForbidden, api.Denied)
		return false
	}

	return true
}

func getUser(context *gin.Context) (store.User, bool) {
	if instance.SingleUser {
		return instance.User(instance.InitialUser), true
	}
	secret := context.GetHeader("secret")
	info := instance.SecretLookup(secret)
	if info.Secret != secret || info.Snowflake == "" {
		return store.User{}, false
	}
	return info, true
}
